package edu.berkeley.sbp;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;

/** an element which matches exactly one input token */
public abstract class Atom<T extends Token> extends Element implements Topology<T> {

    protected abstract Topology<T> top();

    void reachable(HashSet<Sequence.Position> h) { /* do-nothing */ }

    public Topology toAtom() { return dup(); }

    /** equality is based on the underlying <tt>Topology</tt> */
    public int hashCode() { return top().hashCode(); }
    
    /** equality is based on the underlying <tt>Topology</tt> */
    public boolean equals(Object o) { return o != null && o instanceof Atom && ((Atom)o).top().equals(top()); }

    /** declared abstract to force subclasses to override it in a useful manner */
    public abstract String toString();

    // Topology Thunks //////////////////////////////////////////////////////////////////////////////

    public void              add(Topology<T> t)         { top().add(t); }
    public void              add(T t)                   { top().add(t); }
    public void              remove(Topology<T> t)      { top().remove(t); }
    public void              remove(T t)                { top().remove(t); }
    public Topology<T>       dup()                      { return top().dup(); }
    public boolean           contains(T v)              { return top().contains(v); }
    public Topology<T>       fresh()                    { return top().fresh(); }
    public Topology<T>       intersect(Topology<T> t)   { return top().intersect(t); }
    public Topology<T>       minus(Topology<T> t)       { return top().minus(t); }
    public Topology<T>       union(Topology<T> t)       { return top().union(t); }
    public Topology<T>       complement()               { return top().complement(); }
    public boolean           disjoint(Topology<T> t)    { return top().disjoint(t); }
    public boolean           containsAll(Topology<T> t) { return top().containsAll(t); }

    // Subclasses //////////////////////////////////////////////////////////////////////////////

    public static class Infer<T extends Token> extends Atom<T> {
        private final Element e;
        public Infer(Element e) { this.e = e; }
        public Topology<T> top() { return (Topology<T>)e.toAtom(); }
        public String toString() { return e.toString(); }
    }
    
    public static class Invert<T extends Token> extends Atom<T> {
        private final Atom<T> a;
        public Invert(Atom<T> a) { this.a = a; }
        public Topology<T> top() { return ((Topology<T>)a.top()).complement(); }
        public String toString() { return "~"+a; }
    }
    
}

